"""
Nodes that modify inputs to obtain outputs but
depending on a 't' time moment to adjust it to the
time of the video in which the input (a frame of a
video) is being edited. A movement effect is not 
edited the same when we are at the begining of the
video effect than when we are at the end.
"""
from yta_editor_nodes.utils import instantiate_cpu_and_gpu_processors
from yta_editor_nodes.processor import _NodeProcessor
from typing import Union


class _VideoNodeProcessor(_NodeProcessor):
    """
    *Abstract class*
    
    *Singleton class*

    *For internal use only*

    *This class must be inherited by the specific
    implementation of some effect that will be done by
    CPU or GPU (at least one of the options)*

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU, but for 
    video frames, including a `t` time moment parameter
    when processing.
    """

    def __init__(
        self,
        video_node_processor_cpu: Union['_VideoNodeProcessorCPU', None] = None,
        video_node_processor_gpu: Union['_VideoNodeProcessorGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        """
        The `video_node_processor_cpu` and `video_node_processor_gpu`
        have to be set by the developer when building the specific
        classes, but the `do_use_gpu` boolean flag will be set
        by the user when instantiating the class to choose 
        between GPU and CPU.
        """
        # TODO: Validate that are '_VideoNodeProcessorCPU'
        # and '_VideoNodeProcessorGPU' subclasses
        super().__init__(
            node_processor_cpu = video_node_processor_cpu,
            node_processor_gpu = video_node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
        self,
        video_node_processor_cpu: Union['_VideoNodeProcessorCPU', None] = None,
        video_node_processor_gpu: Union['_VideoNodeProcessorGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        super().__reinit__(
            node_processor_cpu = video_node_processor_cpu,
            node_processor_gpu = video_node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def process(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        t: float,
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        return self._processor.process(
            input = input,
            t = t,
            **kwargs
        )
    
class BreathingFrameVideoNodeProcessor(_VideoNodeProcessor):
    """
    The frame but as if it was breathing.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        zoom: float = 0.05
    ):
        # Dynamic way to import it
        video_node_processor_cpu, video_node_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'BreathingFrameVideoNodeProcessor',
            cpu_module_path = 'yta_editor_nodes_cpu.processor.video',
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video',
            gpu_kwargs = {
                'opengl_context': None,
                # TODO: Do not hardcode please...
                'output_size': (1920, 1080),
                'zoom': zoom
            }
        )

        super().__init__(
            video_node_processor_cpu = video_node_processor_cpu,
            video_node_processor_gpu = video_node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
        self,
        do_use_gpu: bool = True,
        zoom: Union[float, None] = None
    ):
        # Dynamic way to import it
        video_node_processor_cpu, video_node_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'BreathingFrameVideoNodeProcessor',
            cpu_module_path = 'yta_editor_nodes_cpu.processor.video',
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video',
            gpu_kwargs = {
                'opengl_context': None,
                # We don't want to change the output
                'output_size': None,
                'zoom': zoom
            }
        )

        super().__init__(
            video_node_processor_cpu = video_node_processor_cpu,
            video_node_processor_gpu = video_node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

class WavingFramesVideoNodeProcessor(_VideoNodeProcessor):
    """
    A video frame that is moving like a wave.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        amplitude: float = 0.05,
        frequency: float = 10.0,
        speed: float = 2.0,
    ):
        # Dynamic way to import it
        video_node_processor_cpu, video_node_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'WavingFramesVideoNodeProcessor',
            cpu_module_path = 'yta_editor_nodes_cpu.processor.video',
            cpu_kwargs = {
                'amplitude': amplitude,
                'frequency': frequency,
                'speed': speed
            },
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video',
            gpu_kwargs = {
                'opengl_context': None,
                # TODO: Do not hardcode please...
                'output_size': (1920, 1080),
                'amplitude': amplitude,
                'frequency': frequency,
                'speed': speed
            }
        )

        super().__init__(
            video_node_processor_cpu = video_node_processor_cpu,
            video_node_processor_gpu = video_node_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
        self,
        do_use_gpu: bool = True,
        amplitude: Union[float, None] = None,
        frequency: Union[float, None] = None,
        speed: Union[float, None] = None,
    ):
        # Dynamic way to import it
        video_node_processor_cpu, video_node_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'WavingFramesVideoNodeProcessor',
            cpu_module_path = 'yta_editor_nodes_cpu.processor.video',
            cpu_kwargs = {
                'amplitude': amplitude,
                'frequency': frequency,
                'speed': speed
            },
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video',
            gpu_kwargs = {
                'opengl_context': None,
                # We don't want to change the output
                'output_size': None,
                'amplitude': amplitude,
                'frequency': frequency,
                'speed': speed
            }
        )

        super().__init__(
            # The CPU doesn't need to be reset
            video_node_processor_cpu = video_node_processor_cpu,
            video_node_processor_gpu = video_node_processor_gpu,
            do_use_gpu = do_use_gpu
        )


"""
The transitions module, with all the classes that
are able to process transitions by using CPU, GPU
or both.

All the classes here will have an instance of the
specific CPU and/or GPU class that is able to run
the code by using either CPU or GPU. The user can
choose between GPU and CPU and that option will be
considered (only if available).

Note for the developer:
A class must have, at least, one specific 
processor (GPU is prefered).

TODO: This module doesn't use 't' but 'progress'
so it is not a child of 'processor.video', maybe
we should move it to be 'processor.transitions'
instead of 'processor.video.transitions'... (?)
"""
from yta_editor_nodes.processor import _ProcessorGPUAndCPU
from yta_editor_nodes.utils import instantiate_cpu_and_gpu_processors
from typing import Union


class _TransitionProcessor(_ProcessorGPUAndCPU):
    """
    *Abstract class*
    
    *Singleton class*

    *For internal use only*

    *This class must be inherited by the specific
    implementation of some transition that will be
    done by CPU or GPU (at least one of the options)*

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def __init__(
        self,
        transition_processor_cpu: Union['_TransitionProcessorCPU', None] = None,
        transition_processor_gpu: Union['_TransitionProcessorGPU', None] = None,
        do_use_gpu: bool = True,
    ):
        """
        The `transition_processor_cpu` and
        `transition_processor_gpu` have to be set by the
        developer when building the specific classes, but
        the `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU.
        """
        super().__init__(
            processor_cpu = transition_processor_cpu,
            processor_gpu = transition_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def process(
        self,
        first_input: Union['moderngl.Texture', 'np.ndarray'],
        second_input: Union['moderngl.Texture', 'np.ndarray'],
        progress: float
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Process the transition between the given `first_input`
        and `second_input`, with GPU or CPU according to the
        internal flag.
        """
        return self._processor.process(
            first_input = first_input,
            second_input = second_input,
            progress = progress,
        )
    
# Specific implementations here below:
class SlideTransitionProcessor(_TransitionProcessor):
    """
    A transition in which the frames goes from one
    side to the other, disappearing the first one
    and appearing the second one.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080)
    ):
        """
        The `transition_processor_cpu` and
        `transition_processor_gpu` have to be set by the
        developer when building the specific classes, but
        the `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU.
        """
        # Dynamic way to import it
        transition_processor_cpu, transition_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'SlideTransitionProcessor',
            cpu_module_path = 'yta_editor_nodes_cpu.processor.video.transitions',
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video.transitions',
            gpu_kwargs = {
                'opengl_context': None,
                'output_size': output_size,
            }
        )

        super().__init__(
            transition_processor_cpu = transition_processor_cpu,
            transition_processor_gpu = transition_processor_gpu,
            do_use_gpu = do_use_gpu
        )

class CrossfadeTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos,
    transforming the first one into the second one.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080)
    ):
        """
        The `transition_processor_cpu` and
        `transition_processor_gpu` have to be set by the
        developer when building the specific classes, but
        the `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU.
        """
        # Dynamic way to import it
        transition_processor_cpu, transition_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'CrossfadeTransitionProcessor',
            cpu_module_path = None,
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video.transitions',
            gpu_kwargs = {
                'opengl_context': None,
                'output_size': output_size,
            }
        )

        super().__init__(
            transition_processor_cpu = transition_processor_cpu,
            transition_processor_gpu = transition_processor_gpu,
            do_use_gpu = do_use_gpu
        )

class DistortedCrossfadeTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos,
    transforming the first one into the second one
    with a distortion in between.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080)
    ):
        """
        The `transition_processor_cpu` and
        `transition_processor_gpu` have to be set by the
        developer when building the specific classes, but
        the `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU.
        """
        # Dynamic way to import it
        transition_processor_cpu, transition_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'DistortedCrossfadeTransitionProcessor',
            cpu_module_path = None,
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video.transitions',
            gpu_kwargs = {
                'opengl_context': None,
                'output_size': output_size,
            }
        )

        super().__init__(
            transition_processor_cpu = transition_processor_cpu,
            transition_processor_gpu = transition_processor_gpu,
            do_use_gpu = do_use_gpu
        )

class AlphaPediaMaskTransitionProcessor(_TransitionProcessor):
    """
    A transition made by using a custom mask to
    join the 2 videos. This mask is specifically
    obtained from the AlphaPediaYT channel in which
    we upload specific masking videos.

    Both videos will be placed occupying the whole
    scene, just overlapping by using the transition
    video mask, but not moving the frame through 
    the screen like other classes do (like the
    FallingBars).

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080),
    ):
        """
        The `transition_processor_cpu` and
        `transition_processor_gpu` have to be set by the
        developer when building the specific classes, but
        the `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU.
        """
        # Dynamic way to import it
        transition_processor_cpu, transition_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'AlphaPediaMaskTransitionProcessor',
            cpu_module_path = None,
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video.transitions',
            gpu_kwargs = {
                'opengl_context': None,
                'output_size': output_size,
            }
        )

        super().__init__(
            transition_processor_cpu = transition_processor_cpu,
            transition_processor_gpu = transition_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def process(
        self,
        first_input: Union['moderngl.Texture', 'np.ndarray'],
        second_input: Union['moderngl.Texture', 'np.ndarray'],
        mask_input: Union['moderngl.Texture', 'np.ndarray'],
        progress: float
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Process the transition between the given `first_input`
        and `second_input`, with GPU or CPU according to the
        internal flag.
        """
        return self._processor.process(
            first_input = first_input,
            second_input = second_input,
            mask_input = mask_input,
            progress = progress,
        )
    
class CircleOpeningTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos in
    which the frames are mixed by generating a circle
    that grows from the middle to end fitting the 
    whole screen.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080),
        border_smoothness: float = 0.02
    ):
        """
        The `transition_processor_cpu` and
        `transition_processor_gpu` have to be set by the
        developer when building the specific classes, but
        the `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU.
        """
        # Dynamic way to import it
        transition_processor_cpu, transition_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'CircleOpeningTransitionProcessor',
            cpu_module_path = None,
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video.transitions',
            gpu_kwargs = {
                'opengl_context': None,
                'output_size': output_size,
                'border_smoothness': border_smoothness
            }
        )

        super().__init__(
            transition_processor_cpu = transition_processor_cpu,
            transition_processor_gpu = transition_processor_gpu,
            do_use_gpu = do_use_gpu
        )

    def __reinit__(
        self,
        do_use_gpu: bool = True,
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080),
        border_smoothness: float = 0.02
    ):
        # TODO: Review and refactor this, I'm using the
        # code twice and repeating it in all the classes
        # Dynamic way to import it
        transition_processor_cpu, transition_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'CircleOpeningTransitionProcessor',
            cpu_module_path = None,
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video.transitions',
            gpu_kwargs = {
                'opengl_context': None,
                'output_size': output_size,
                'border_smoothness': border_smoothness
            }
        )

        super().__reinit__(
            processor_cpu = transition_processor_cpu,
            processor_gpu = transition_processor_gpu,
            do_use_gpu = do_use_gpu
        )

class CircleClosingTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos in
    which the frames are mixed by generating a circle
    that grows from the middle to end fitting the 
    whole screen.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080),
        border_smoothness: float = 0.02
    ):
        """
        The `transition_processor_cpu` and
        `transition_processor_gpu` have to be set by the
        developer when building the specific classes, but
        the `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU.
        """
        # Dynamic way to import it
        transition_processor_cpu, transition_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'CircleClosingTransitionProcessor',
            cpu_module_path = None,
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video.transitions',
            gpu_kwargs = {
                'opengl_context': None,
                'output_size': output_size,
                'border_smoothness': border_smoothness
            }
        )

        super().__init__(
            transition_processor_cpu = transition_processor_cpu,
            transition_processor_gpu = transition_processor_gpu,
            do_use_gpu = do_use_gpu
        )

class BarsFallingTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos in which
    a set of bars fall with the first video to let the
    second one be seen.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.
    """

    def __init__(
        self,
        do_use_gpu: bool = True,
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080),
        number_of_bars: int = 30,
        amplitude: float = 2,
        noise: float = 0.1,
        frequency: float = 0.5,
        drip_scale: float = 0.5,
    ):
        """
        The `transition_processor_cpu` and
        `transition_processor_gpu` have to be set by the
        developer when building the specific classes, but
        the `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU.
        """
        # Dynamic way to import it
        transition_processor_cpu, transition_processor_gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'BarsFallingTransitionProcessor',
            cpu_module_path = None,
            cpu_kwargs = {},
            gpu_module_path = 'yta_editor_nodes_gpu.processor.video.transitions',
            gpu_kwargs = {
                'opengl_context': None,
                'output_size': output_size,
                'number_of_bars': number_of_bars,
                'amplitude': amplitude,
                'noise': noise,
                'frequency': frequency,
                'drip_scale': drip_scale
            }
        )

        super().__init__(
            transition_processor_cpu = transition_processor_cpu,
            transition_processor_gpu = transition_processor_gpu,
            do_use_gpu = do_use_gpu
        )


from yta_validation.parameter import ParameterValidator
from yta_programming.singleton import SingletonABCMeta
from abc import abstractmethod
from typing import Union


class _ProcessorGPUAndCPU(metaclass = SingletonABCMeta):
    """
    *Abstract class*

    *For internal use only*

    Abstract class to share the common behaviour of
    being able to handle a process with a GPU and/or
    a CPU processor, chosen by the user.
    """

    @property
    def is_gpu_available(
        self
    ) -> bool:
        """
        Boolean flag to indicate if the GPU is available
        or not, that means that the processor that uses
        GPU is set.
        """
        return self._processor_gpu is not None
    
    @property
    def is_cpu_available(
        self
    ) -> bool:
        """
        Boolean flag to indicate if the CPU is available
        or not, that means that the processor that uses
        CPU is set.
        """
        return self._processor_cpu is not None
    
    @property
    def _processor(
        self
    ) -> Union['_ProcessorCPU', '_ProcessorGPU']:
        """
        *For internal use only*
        
        Get the processor that must be applied to process
        the inputs according to the internal flag that
        indicates if we want to use GPU or CPU and also
        depending on the availability of these classes.
        """
        return (
            (
                # Prefer GPU if available
                self._processor_gpu or
                self._processor_cpu
            ) if self._do_use_gpu else (
                # Prefer CPU if available
                self._processor_cpu or
                self._processor_gpu
            )
        )
    
    def __init__(
        self,
        processor_cpu: Union['_ProcessorGPU', None] = None,
        processor_gpu: Union['_ProcessorCPU', None] = None,
        do_use_gpu: bool = True,
    ):
        """
        The `processor_cpu` and
        `processor_gpu` have to be set by the
        developer when building the specific classes, but
        the `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU.
        """
        ParameterValidator.validate_mandatory_bool('do_use_gpu', do_use_gpu)

        if (
            processor_cpu is None and
            processor_gpu is None
        ):
            raise Exception('No node processor provided. At least one node processor is needed.')

        self._processor_cpu: Union['_ProcessorCPU', None] = processor_cpu
        """
        The transition processor that is able to do the
        processing by using the CPU. If it is None we cannot
        process it with CPU.
        """
        self._processor_gpu: Union['_ProcessorGPU', None] = processor_gpu
        """
        The transition processor that is able to do the
        processing by using the GPU. If it is None we cannot
        process it with GPU.
        """
        self._do_use_gpu: bool = do_use_gpu
        """
        Internal flag to indicate if we should use GPU,
        if True, or CPU if False.
        """

    def __reinit__(
        self,
        processor_cpu: Union['_ProcessorGPU', None] = None,
        processor_gpu: Union['_ProcessorCPU', None] = None,
        do_use_gpu: bool = True,
    ):
        if (
            processor_cpu is None and
            processor_gpu is None
        ):
            raise Exception('No node processor provided. At least one node processor is needed.')

        self._processor_cpu: Union['_ProcessorCPU', None] = processor_cpu
        self._processor_gpu: Union['_ProcessorGPU', None] = processor_gpu
        self._do_use_gpu: bool = do_use_gpu

    def use_gpu(
        self
    ) -> '_ProcessorGPUAndCPU':
        """
        Set the internal flag to use GPU if available.
        """
        self._do_use_gpu = True

        return self

    def use_cpu(
        self
    ) -> '_ProcessorGPUAndCPU':
        """
        Set the internal flag to use CPU if available.
        """
        self._do_use_gpu = False

        return self

    @abstractmethod
    def process(
        self,
        input: Union['moderngl.Texture', 'np.ndarray']
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Process the provided 'input' with GPU or CPU 
        according to the internal flag.
        """
        pass


from yta_validation import PythonValidator
from typing import Union

import importlib


def is_gpu_available(
) -> bool:
    """
    Check if the GPU module is installed or not.
    """
    return PythonValidator.is_dependency_installed('yta_editor_nodes_gpu')

def is_cpu_available(
) -> bool:
    """
    Check if the CPU module is installed or not.
    """
    return PythonValidator.is_dependency_installed('yta_editor_nodes_cpu')


def instantiate_cpu_and_gpu_processors(
    class_name: str,
    cpu_module_path: Union[str, None] = 'yta_editor_nodes_cpu.processor.blender',
    cpu_kwargs: Union[dict, None] = None,
    gpu_module_path: Union[str, None] = 'yta_editor_nodes_gpu.processor.blender',
    gpu_kwargs: Union[dict, None] = None
):
    """
    Obtain instances of the classes with the given
    `class_name` (with CPU or GPU suffix) from the
    given modules, using the also given `cpu_kwargs`
    and `gpu_kwargs`.

    If `cpu_module_path` or `gpu_module_path` is
    None, the class will not be imported. Useful 
    when you only have one service available (GPU
    or CPU only).

    Example:
        ```
        cpu, gpu = instantiate_cpu_and_gpu_processors(
            class_name = 'MixBlender',
            module_path_cpu = 'yta_editor_nodes_cpu.processor.blender',
            module_path_gpu = 'yta_editor_nodes_gpu.processor.blender',
            cpu_kwargs = {},
            gpu_kwargs = {}
        )
        ```

    This example will try to instantiate the
    `MixBlenderCPU` and `MixBlenderGPU` classes from
    those modules.
    """
    processor_cpu = None
    do_has_cpu_dependency_installed = is_cpu_available()
    processor_gpu = None
    do_has_gpu_dependency_installed = is_gpu_available()

    if (
        not do_has_cpu_dependency_installed and
        not do_has_gpu_dependency_installed
    ):
        raise Exception('No CPU nor GPU dependency was found. Install the CPU dependency with the "pip install yta_editor_nodes[yta_editor_nodes_cpu]" command, or the GPU dependency with "pip install yta_editor_nodes[yta_editor_nodes_gpu]".')

    if (
        cpu_module_path is not None and
        do_has_cpu_dependency_installed
    ):
        try:
            print(f'Importing {cpu_module_path} {class_name}CPU')
            cpu_module = importlib.import_module(cpu_module_path)
            processor_cpu = getattr(cpu_module, f'{class_name}CPU')(**(cpu_kwargs or {}))
        except (ImportError, AttributeError):
            pass

    if (
        gpu_module_path is not None and
        do_has_gpu_dependency_installed
    ):
        try:
            print(f'Importing {gpu_module_path} {class_name}GPU')
            gpu_module = importlib.import_module(gpu_module_path)
            processor_gpu = getattr(gpu_module, f'{class_name}GPU')(**(gpu_kwargs or {}))
        except (ImportError, AttributeError):
            pass

    if (
        processor_cpu is None and
        processor_gpu is None
    ):
        raise Exception(f'No CPU nor GPU module found for the "{class_name}" class.')

    return (
        processor_cpu,
        processor_gpu
    )



"""
Our awesome editor module in which we
have all the classes that interact with
it and make it possible.

This is the nodes module, in which we have
all the classes that make the concept work.
"""
from abc import ABC, abstractmethod
from typing import Union


class Node(ABC):
    """
    *Abstract class*

    *This class has to be inherited by any class that
    is able to handle some input to obtain an output
    as a result*

    The abstract class of a Node, which is the entity
    that is able to process some input to return an
    output that can be sent to the next node.
    """

    @abstractmethod
    def process(
        self,
        input: Union['np.ndarray', 'moderngl.Texture'],
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the input and obtain a result as the output.
        """
        pass


"""
Note for the developer:

A guide to the different types of nodes we will have
when imitating DaVinci Resolve.

Node (abstracto)
├── ProcessorNode (procesa un solo flujo)
│   ├── EffectNode (filtros, LUTs, shaders…)
│   └── TransitionNode (mezcla entre dos entradas)
│
├── CompositeNode (combina múltiples flujos)
│   ├── SerialNode (uno tras otro)
│   ├── ParallelNode (ramas simultáneas)
│   └── LayerNode (capas con blending o máscaras)
│
└── GroupNode (contiene una mini subred de nodos)

"""
